open Mcic;;
open Env;;
open Typedef;;
open Emit;;
open StringConst;;

exception NotGlobal;;
exception NotSupport;;
exception Redefined of string;;
exception TypeNotMatch;;
exception NotConstant;;
exception FuncUnmatchType;;

let get_ident_type id = ((current())#get id)#typ

type cpu_2op = ADDL | SUBL | IMULL | DIVL | MOVL | MOVB | LEAL | CMPL
	       | TESTL ;;
type cpu_1op = JMP | Je | Jne | PUSHL | CALL | POPL | SETne | SETe 
	       | SETle | SETl | SETge | SETg;;

type address =
    Register of int
  | WRegister of int
  | BRegister of int
  | Esp 
  | Constant of int
  | Localvar of int
  | Label of string
  | IndirectAdr of address
  | OffSetAdr of address * address
  | LeaAdr of address * int * int

let rec unuse_address adr =
  match adr with
      Register r -> unuse_reg r
    | IndirectAdr adr -> unuse_address adr
    | OffSetAdr (a1,a2) -> unuse_address a1; unuse_address a2
    | LeaAdr (a1,multi,off) -> unuse_address a1
    | _ -> ();;


(*
  EAX = 0
  EBX
  ECX
  EDX
  ESI
  EDI = 5 *)

let reg_to_breg reg =
  match reg with
      Register num -> BRegister num
    | _ -> reg;;

let reg_name_tbl = [|"%eax";"%ebx";"%ecx";"%edx";"%esi";"%edi"|];;
let reg_name_tblw = [|"%ax";"%bx";"%cx";"%dx";"%si";"%di"|];;
let reg_name_tblb = [|"%al";"%bl";"%cl";"%dl"|];;

let reg_name num = reg_name_tbl.(num);;
let wreg_name num = reg_name_tblw.(num);;
let breg_name num = reg_name_tblb.(num);;

let num_name num = "$"^(string_of_int num);;

let op2_name op =
  match op with
      ADDL -> "addl"
    | SUBL -> "subl"
    | IMULL -> "imull"
    | DIVL -> "divl"
    | MOVL -> "movl"
    | MOVB -> "movb"
    | LEAL -> "leal"
    | CMPL -> "cmpl"
    | TESTL -> "testl";;

let op1_name op =
  match op with
      JMP -> "jmp"
    | Je -> "je"
    | Jne -> "jne"
    | PUSHL -> "pushl"
    | POPL -> "popl"
    | CALL -> "call"
    | SETe -> "sete"
    | SETne -> "setne"
    | SETle -> "setle"
    | SETl -> "setl"
    | SETge -> "setge"
    | SETg -> "setg"

let rec address_name a =
  match a with
      Register r -> reg_name r
    | WRegister r -> wreg_name r
    | BRegister r -> breg_name r
    | Esp -> "%esp"
    | Constant i -> num_name i
    | Localvar i -> (string_of_int i)^"(%ebp)"
    | Label l -> l	
    | IndirectAdr a -> "("^(address_name a)^")"
    | OffSetAdr(x,y) -> (address_name y)^"("^(address_name x)^")"
    | LeaAdr( base, multi, off ) ->
	(string_of_int off)^"(,"^(address_name base)^","^
	(string_of_int multi)^")"

(*  gas 촹 mov eax, 1 -> mov $1, %eax*)
let emit_2op op r1 r2 = 
  emit_string ("\t"^(op2_name op)^"\t"^(address_name r2)^", "^(address_name r1));;

let emit_1op op r =
  emit_string ("\t"^(op1_name op)^"\t"^(address_name r));;


type binary_op = { op: int->int->int; commutative: bool; gen_func: address->address->address }

let gen_arith_code cpu_op l_adr r_adr =
  emit_2op cpu_op l_adr r_adr;
  unuse_address r_adr;
  l_adr;;

let gen_cmp_code cpu_op_arg l_adr_arg r_adr_arg =
  let (cpu_op,l_adr,r_adr) = 
    match cpu_op_arg with
	SETg -> ( SETle, r_adr_arg, l_adr_arg )
      | SETge -> ( SETl, r_adr_arg, l_adr_arg )
      | _ -> ( cpu_op_arg, l_adr_arg, r_adr_arg )
  in
    emit_2op CMPL l_adr r_adr;
    unuse_address r_adr;
    emit_1op cpu_op (reg_to_breg l_adr);
    l_adr;;

let binary_op_table = 
  Util.create_hashtbl 133 
    [("+",{op=(+);commutative=true;gen_func= gen_arith_code ADDL};);
     ("*",{op=( * );commutative=true;gen_func= gen_arith_code IMULL});
     ("-",{op=(-);commutative=false;gen_func= gen_arith_code SUBL});
     ("/",{op=(/);commutative=false;gen_func= gen_arith_code DIVL});
     ("!=",{op=(fun x y -> if x!=y then 1 else 0);commutative=true;gen_func= gen_cmp_code SETne});
     ("==",{op=(fun x y -> if x =y then 1 else 0);commutative=true;gen_func= gen_cmp_code SETe});
     ("<" ,{op=(fun x y -> if x <y then 1 else 0);commutative=false;gen_func= gen_cmp_code SETl});
     (">" ,{op=(fun x y -> if x >y then 1 else 0);commutative=false;gen_func= gen_cmp_code SETg});
     ("<=",{op=(fun x y -> if x<=y then 1 else 0);commutative=false;gen_func= gen_cmp_code SETle});
     (">=",{op=(fun x y -> if x>=y then 1 else 0);commutative=false;gen_func= gen_cmp_code SETge});
    ];;

let error_cnt = ref 0;;

let emit_char_int reg = 
  emit_string ("\tmovsbl\t"^(breg_name reg)^", "^(reg_name reg)^"\n")

let rec is_constant expr = 
  match expr with
      Int_Constant i -> true
    | Funcall (func,arg) ->
	(match func with
	     ID id -> is_binary_constatnt id arg
	   | _ -> false)
    | _ -> false
and is_binary_constatnt func arg = 
  let l = is_constant (List.nth arg 0) in
    if l = false then
      false
    else
      let r = is_constant (List.nth arg 1)
      in
	Hashtbl.mem binary_op_table func
	      
let rec get_constant expr =
  match expr with 
      Int_Constant i -> i
    | Funcall( func, arg )->
	(match func with
	     ID id ->
	       if Hashtbl.mem binary_op_table id then
		 get_binary_constant id arg
	       else
		 raise NotConstant
	   | _ ->
	       raise NotConstant)

    | _ -> raise NotConstant
and get_binary_constant funcname arg  =
  let l = get_constant (List.nth arg 0)
  and r = get_constant (List.nth arg 1)
  in
    (Hashtbl.find binary_op_table funcname).op l r

let type_expr_to_type = Typedef.type_expr_to_type is_constant get_constant 

let rec check_funcall_type args_typ args = 
  match args_typ with
      hd_1::tl_1 -> 
	if hd_1 = VarArgType then
	  true
	else
	  begin
	    match args with 
		hd_2::tl_2 -> 
		  if aprox_type hd_1 (calc_expr_type hd_2) then
		    check_funcall_type tl_1 tl_2
		  else
		    ( prerr_endline "η㤦"; raise FuncUnmatchType )
	      | [] ->
		  prerr_endline "ο㤦";
		  raise FuncUnmatchType
	  end

    | [] ->
	match args with 

	    hd::tl ->
 	      prerr_endline "ο㤦";
	      raise FuncUnmatchType
	  | [] -> true;; 
	    
let id_to_address char_hook id =
  try 
    let tbl = find_id id in
    let var = tbl#get id in
    let adr = match tbl#get_type with
	LocalTable|FunctionTable ->
	  Localvar( var#get_offset() )
      | _ ->
	  Label( id )
    in
      match var#typ with
	  IntType -> 
	    adr
	| CharType ->
	    char_hook adr
	| Pointer_type p ->
	    adr
	| Array_type (arr,off)->
	    let ret = usable_reg()
	    in
	      use_reg ret;
	      emit_2op LEAL (Register ret) adr ;
	      unuse_address adr;
	      (Register ret)
		
	| _ ->
	    prerr_endline "not integer type";
	    raise TypeNotMatch
	      
      with Not_found->
	prerr_endline (id^"is not declared");
	raise Not_found;;

let char_to_int adr =
  let reg = usable_reg() 
  in
    use_reg reg;
    emit_2op MOVB (BRegister(reg)) adr;
    emit_char_int reg;
    unuse_address adr;
    Register( reg );;

let id_to_int_address = id_to_address char_to_int ;;

let rec gen_expr expr =
  match expr with
      Int_Constant num -> Constant(num)
    | Str_Constant str -> Label("$"^gen_str_const str)
    | ID id -> id_to_int_address id
    | Funcall( func, args ) ->
	gen_funcall func args
    | Asign_Expr( dst, src ) ->
	gen_asign dst src ;
    | Indirect_Ref address ->
	gen_indirect address
    | Array_Ref (array,off) ->
	gen_array_ref array off
    | _ ->
	prerr_endline "not support in gen_expr";
	raise NotSupport;

and gen_array_ref array off =
  let a_reg = gen_expr_reg array
  and off_adr = gen_expr_reg off
  and elem_size = get_typ_size (get_ref_type array)
  in
    emit_2op LEAL off_adr (LeaAdr (off_adr,elem_size,0));
    emit_2op ADDL a_reg off_adr;
    unuse_address off_adr;
    (IndirectAdr a_reg)

and gen_indirect adr =
  let reg = gen_expr_reg adr
  in
    match calc_expr_type adr with
	CharType -> 
	  emit_2op MOVB reg (IndirectAdr reg);
	  reg;
      | _ ->
	  emit_2op MOVL reg (IndirectAdr reg);
	  reg;

and gen_arrayref_lval arr off =
  gen_array_ref arr off

and gen_expr_lval expr =
  match expr with
      ID id -> id_to_address (fun x->x) id
    | Array_Ref ( arr, off ) -> gen_arrayref_lval arr off
    | _ ->
	prerr_endline "ͤʤΤ";
	raise NotSupport;

and gen_expr_reg expr =			(* ̤ɬ쥸 *)
  let e = gen_expr expr in
  let gen_mov = 
    let reg = usable_reg()
    in
      use_reg reg;
      emit_2op MOVL (Register reg) e;
      Register reg
  in
    match e with
	Label l -> gen_mov
      | Constant num -> gen_mov
      | Localvar off -> gen_mov
      | IndirectAdr adr -> 
	  let ret = gen_mov in
	    unuse_address adr;
	    ret
      | _ ->
	  e

and expr_to_address expr =
  match expr with
      Int_Constant i -> (Constant i)
    | ID id -> id_to_int_address id
    | _ -> raise (InternalError("expr_to_address"))

and gen_asign dst src =
  let r_reg = gen_expr_reg src 
  and l_reg = gen_expr_lval dst 
  in
    (
      match calc_expr_type dst with
	  IntType -> emit_2op MOVL l_reg r_reg;
	| Pointer_type p -> emit_2op MOVL l_reg r_reg;
	| CharType -> emit_2op MOVB (reg_to_breg l_reg) (reg_to_breg r_reg)
	| _ -> prerr_endline "gen_asign η"; raise TypeNotMatch
    );

    emit_comment "asing ҤȤʬ";
	    
    unuse_address r_reg;
    l_reg

	
and gen_binary opname args =
  let binary_op = Hashtbl.find binary_op_table opname in
  let commutative = binary_op.commutative
  and gen_func = binary_op.gen_func
  and constant_or_expr expr =
    if is_constant expr then
      Int_Constant( get_constant expr )
    else
      expr
  in
  let l = constant_or_expr (List.nth args 0)
  and r = constant_or_expr (List.nth args 1)
  and rl l r_reg =
    let l_reg = usable_reg() in
    let l_adr = Register( l_reg )
    in
      emit_2op MOVL l_adr l;
      use_reg l_reg;
      gen_func l_adr r_reg;

  in
    match (l,r) with
	(Int_Constant i0, Int_Constant i1 ) ->
	  Constant (get_binary_constant opname [l;r]);
      | (Int_Constant i,_) -> 
	  let r_reg = gen_expr_reg r
	  in
	    if commutative then
	      gen_func r_reg (Constant i)
	    else 
	      rl (Constant i) r_reg

      | (ID id,_) ->
	  let r_reg = gen_expr_reg r
	  in
	    if commutative then
	      let l_reg = id_to_int_address id
	      in
		gen_func r_reg l_reg;
	    else
	      rl (id_to_int_address id) r_reg
      | (_,(Int_Constant i)) ->
	  let l_reg = gen_expr_reg l
	  in
	    gen_func l_reg (Constant i);

      | (_,(ID id)) ->
	  let l_reg = gen_expr_reg l
	  and r_reg = id_to_int_address id
	  in
	    gen_func l_reg r_reg;
      | _ ->
	  let r_reg = gen_expr_reg r 
	  and l_reg = gen_expr_reg l
	  in
	    gen_func l_reg r_reg


and gen_funcall_id func_id args =
  try 
    let tbl = find_id func_id in
    let func_var = tbl#get func_id in
    let typ = func_var#typ
    in
      begin
	match typ with
	    FuncType( ret, args_typ ) ->
	      if check_funcall_type args_typ args then
		begin
		    List.iter 
		      (fun arg_exp ->
			 let reg = gen_expr arg_exp
			 in
			   emit_1op PUSHL reg;
			   unuse_address reg) (List.rev args);
		  emit_1op CALL (Label func_id);
		  emit_2op ADDL Esp (Constant (List.length args*4));
		  (Register 0);
		end
	      else
		(prerr_endline "ؿη㤤Ŭ";
		 raise (ApplyToNotFunction(func_id)))
		
	  | _ ->
	      prerr_endline (func_id^" ϴؿʤΤˡ");
	      raise (ApplyToNotFunction(func_id))
      end

  with Not_found ->
    prerr_endline (func_id^"ϲ");
    raise Not_found;

and gen_funcall func args =
  match func with 
      ID id ->
	if Hashtbl.mem binary_op_table id then
	  let chklen l = 
	    if List.length l != 2 then
	      prerr_endline "ѱ黻ΰ㤦"
	    else
	      ()
	  in
	    chklen args;
	    gen_binary id args
	else 
	  gen_funcall_id id args
    | _ -> raise NotSupport

and gen_expr_stmt expr =
  match expr with
    | Asign_Expr( e1, e2 ) -> unuse_address(gen_asign e1 e2)
    | ID i -> ()
    | Int_Constant i -> ()
    | Str_Constant s -> ()
    | _ -> unuse_address(gen_expr expr);;

let rec gen_func ( ret, id, args, body ) =
  let is_main_func = if id = "main" then true else false in
    if (current())#get_type != GlobalTable then
      begin
	prerr_endline ("function "^id^" is defined in not global");
	raise NotGlobal;
      end
    else
      begin
	let tbl = (new symbol_table FunctionTable) 
	and exit_label = gen_local_label() 
	and main_ebx = if is_main_func then 4 else 0
	in
	  emit_text (); 
	  emit_globl_label id;
	  set_exit_label exit_label;
	  set_rettype ret;
	  tbl#set_parents( get_env_stack() );
	  push tbl;
	  
	  let off_max = (List.length args) * 4 + 4 in
	    
	  let rec regist_args arg_list =
	    match arg_list with
		hd::tl -> 
		  let var = (new variable (snd hd) (type_expr_to_type (fst hd)) false)
		  in
		    (current())#regist var;
		    let off = regist_args tl 
		    in
		      var#set_offset off;
		      off + 4
	      | [] ->
		  8
	  in
	    ignore(regist_args (List.rev args));
	    emit_make_frame();
	    if is_main_func then (emit_1op PUSHL (Register 1); (current())#inc_size 4) else ();
	    gen_stmt_list body;
	    emit_label exit_label;
	    if is_main_func then emit_1op POPL (Register 1) else ();
	    emit_remove_frame();
	    ignore((current())#align 4);
	    emit_func_end ((current())#get_max_size()+main_ebx);
	    pop();
      end
and gen_stmt stmt = 
  let declare_local_var id t_exp size =
    let c = current()
    and var = new variable id t_exp false in
    let offset = (~-(c#align (if size>4 then size else size))) - 4 in
      if c#mem id then
	raise (Redefined id)
      else
	begin
	  var#set_offset offset;
	  c#inc_size size;
	  c#regist var;
	end
  and gen_cond_expr expr = 
    let cond_val = gen_expr_reg expr
    in
      emit_2op TESTL cond_val cond_val;
      unuse_address cond_val;
  in

    match stmt with
	DeclFunc ( ret, id, args, body ) ->
	  gen_func( (type_expr_to_type ret), id, args, body ) ;
	  ignore();
      | A_Stmt expr ->
	  ignore ( gen_expr_stmt expr );
      | DeclVarStmt(expr,id) -> 
	  let t_exp = type_expr_to_type expr in
	  let size = get_typ_size t_exp in
	    if ((current())#get_type = LocalTable ||
		(current())#get_type = FunctionTable ) then
		declare_local_var id t_exp size
	    else
	      begin
		((current())#expand (get_typ_size t_exp));
		emit_bss();
		emit_align 4;
		emit_globl_label id;
		emit_storage 4;
	      end

      | DeclVarStmtInit(expr,id,init) ->
	  let t_exp = type_expr_to_type expr in
	  let size = get_typ_size t_exp in
	    if ((current())#get_type = LocalTable ||
		(current())#get_type = FunctionTable ) then
	      begin
		declare_local_var id t_exp size;
		unuse_address(gen_asign (ID(id)) init);
	      end
	    else
	      if (is_constant init)=false then
		(prerr_endline "Хѿνͤʤ";
		 raise NotConstant)
	      else 
		begin
		  emit_data();
		  emit_align 4;
		  emit_globl_label id;
		  emit_long (get_constant init);
		end

      | ReturnStmt ( expr ) ->
	  let ret_reg = gen_expr expr 
	  in
	    unuse_address ret_reg;
	    emit_2op MOVL (Register 0) ret_reg;
	    begin
	      match calc_expr_type expr with
		  CharType -> 
		    emit_char_int 0
		| _ ->
		    ()
	    end;

	    emit_1op JMP (Label (get_exit_label()))
      | DeclExtern(ret,id,args) ->
	  ()
      | WhileStmt ( cond, body ) ->
	  let exit_label = gen_local_label ()
	  and loop_top = gen_local_label ()
	  in
	    emit_label loop_top;
	    gen_cond_expr cond;
	    emit_1op Je (Label exit_label);
	    
	    gen_stmt_list body;
	    emit_1op JMP (Label loop_top);
	    emit_label exit_label
      | IfStmt ( cond, true_body, elsif_stmt, false_body ) ->
	  let exit_label = gen_local_label()
	  and cond_val = gen_expr_reg cond in
	  let gen_cond ( cond, body ) =
	    let false_label = gen_local_label()
	    in
	      gen_cond_expr cond;
	      emit_1op Je (Label false_label);
	      
	      gen_stmt_list body;
	      emit_1op JMP (Label exit_label);

	      emit_label false_label
	  in
	    gen_cond (cond,true_body);
	    List.iter gen_cond elsif_stmt;
	    gen_stmt_list false_body;
	    emit_label exit_label

      | ImportStmt ->
	  ()

      | _->
	  prerr_endline "not support statement";
	  raise NotSupport
and gen_stmt_list slist = List.iter gen_stmt slist;;

let pre_analyze_stmt_list analayze_func slist = List.iter analayze_func slist;;

(* 1ѥܤϷ *)
let analyze_type stmt = ();;

let analyze_global_decl stmt = 
  let c = Env.current()
  and arg_decl_to_type x = 
    type_expr_to_type (fst x)
  in
    match stmt with
	DeclFunc (ret,id,args,body) -> 
	  c#regist (new Env.variable id (FuncType(type_expr_to_type ret,
						  (List.map arg_decl_to_type args))) true);
      | DeclExtern (ret,id,args) ->
	  c#regist (new Env.variable id (FuncType(type_expr_to_type ret,
						  (List.map arg_decl_to_type args))) true)
      | DeclVarStmt ( typ, id ) ->
	  c#regist (new Env.variable id (type_expr_to_type typ) true);
      | DeclVarStmtInit ( typ, id, init ) ->
	  c#regist (new Env.variable id (type_expr_to_type typ) true);
      | _ -> ();;

let genCode out program =
  set_output out;
  emit_comment ("# generated by mcc"^"\n");
  gen_stmt_list program;
  emit_rodata();
  flush_str_const(); 
  emit_string "\n";
  emit_flush();
  !error_cnt;;
