%{

open Mcic;;
open Typedef;;

type variable = Env.variable;;

let parse_error err =
  print_endline ((string_of_int !Location.lineno)^" "^err);;

let get_type_expr = Typedef.type_expr_to_type CodeGen.is_constant CodeGen.get_constant;;

%}
%token EOF

%token <int> INT_LITERAL
%token <string> STR_LITERAL

%token HUT /* ^ */

%token PLUS
%token MINUS
%token STAR
%token SLASH
%token SEMI
%token SHARP
%token BEGIN
%token END
%token LOOP
%token WHILE
%token IF
%token ELSE
%token ELSIF
%token THEN
%token DO
%token DONE
%token <string> IDENT
%token COMMA
%token DOT
%token DOTDOTDOT
%token QUOTE
%token GT /* > */
%token LT /* < */
%token AND /* & */
%token ANDAND
%token OR /* | */
%token OROR 
%token EQU /* == */
%token NEQU
%token ASIGN
%token RETURN
%token EXTERN 
%token IMPORT

%token CLASS

%token LPAR /* ( ) */
%token RPAR 

%token LBRACK /* [ ] */
%token RBRACK
%token LBRACE /* { } */
%token RBRACE  

%start program
%type <Mcic.program> program

%right ASIGN

%left NEQU EQU
%left GT LT
%left PLUS MINUS
%left STAR SLASH

%left DOT

%nonassoc UNARY

%nonassoc LPAR RPAR LBRACK RBRACK

%%
;

program:
  global_stmt_list EOF{ $1 }
;

global_stmt_list:
   /* */  { [] }
| global_stmt global_stmt_list { $1::$2 }
;

global_stmt:
  declare_global { $1 }
| import_stmt { $1 }
;

import_stmt:
  IMPORT identifier SEMI { (Env.compile_func())( $2^".mc" ); ImportStmt; }
;

stmt_list:
  /* */          { [] }
  | stmt stmt_list { $1::$2 }
;

stmt:
  expr SEMI { A_Stmt($1) }
  | var_declare SEMI { $1 }
  | IF expr THEN stmt_list elsif_list else_block END { IfStmt( $2, $4, $5, $6 ) }
  | LOOP stmt_list END { LoopStmt( $2 ) }
  | WHILE expr DO stmt_list DONE { WhileStmt( $2, $4 ) }
  | RETURN expr SEMI { ReturnStmt( $2 ) }
  | block	{ $1 }
;

else_block:
  /* */ { [] }
| ELSE stmt_list { $2 }

elsif_list:
  ELSIF expr THEN stmt_list elsif_list
  { ( $2, $4 )::$5 }
| /* */ { [] }

block:
  BEGIN stmt_list END { BlockStmt($2) }
;

declare_global:
  var_declare_global SEMI { $1 }
| func_declare_global { $1 }
/*| var_declare_extern SEMI { $1 }*/
| func_declare_extern SEMI { $1 }
;

identifier:
  IDENT { $1 }

func_declare_global:
  type_expr identifier LPAR comma_arg_decl_list1 RPAR stmt_list END
  { 
    DeclFunc( $1, $2, $4, $6 )
  }

;

func_declare_extern:
  EXTERN type_expr identifier LPAR comma_arg_decl_list1 RPAR 
  {
    DeclExtern( $2, $3, $5 )
  }

comma_arg_decl_list1:
  arg_decl comma_arg_decl_list { $1::$2 }
| /* */  { [] }
;

comma_arg_decl_list:
  COMMA arg_decl comma_arg_decl_list { $2::$3 }
| COMMA DOTDOTDOT { [( ID("..."), "..." )] }
| /* */  { [] }
;

arg_decl:
  type_expr identifier { 
    ($1,$2)
  }
;

var_declare_global:
  type_expr identifier ASIGN expr	{ DeclVarStmtInit( $1, $2, $4 ) }
| type_expr identifier			{ DeclVarStmt( $1, $2 ) }
;

/*var_declare_extern:
  EXTERN type_expr identifier { DeclExtern($2,$3) }
;*/

var_declare:
  var_declare1 ASIGN expr	{ DeclVarStmtInit( fst $1, snd $1, $3 ) }
| var_declare1		{ DeclVarStmt( fst $1, snd $1 ) }
;

/* ( typ, id ) */ 

var_declare1:
  var_declare1 LBRACK expr RBRACK { ( Array_Decl(fst $1,$3), snd $1 ) }
| var_declare_type_ident { $1 }
;

var_declare_type_ident:
  type_expr identifier { ($1,$2) }
;

type_expr:
  type_expr STAR { Expr_Star $1 }
| identifier { ID($1) }
;

expr:
  expr NEQU expr	{ Funcall(ID("!="),[$1;$3]) }
| expr EQU expr		{ Funcall(ID("=="),[$1;$3]) }
| expr GT expr		{ Funcall(ID(">"),[$1;$3]) }
| expr LT expr 		{ Funcall(ID("<"),[$1;$3]) }
| expr PLUS expr 	{ Funcall(ID("+"),[$1;$3]) }
| expr MINUS expr 	{ Funcall(ID("-"),[$1;$3]) }
| expr STAR expr 	{ Funcall(ID("*"),[$1;$3]) }
| expr SLASH expr 	{ Funcall(ID("/"),[$1;$3]) }
| expr ASIGN expr 	{ Asign_Expr( $1, $3 ) }
| unary_expr		{ $1 }  
| LBRACE stmt RBRACE	{ Stmt_Expr $2 }
;

unary_expr:
  STAR unary_expr 	{ Indirect_Ref $2 }  
| postfix		{ $1 }
;

postfix:
  postfix LBRACK expr RBRACK { Mcic.Array_Ref ( $1, $3 ) }
| postfix LPAR comma_expr_list1 RPAR { Funcall( $1, $3 ) }
| primary		{ $1 }
;

primary:
  constant		{ $1 }
| identifier		{ ID( $1 ) }
| term 			{ $1 }
;

term:
| LPAR expr RPAR 	{ $2 }
;

comma_expr_list1:
  /* */				{ [] }
| expr comma_expr_list		{ $1::$2 }
;

comma_expr_list:
  /* */				{[]}
| COMMA expr comma_expr_list 	{ $2::$3 }
;

constant:
  INT_LITERAL { Int_Constant($1) }
| STR_LITERAL { Str_Constant($1) }
;
